unit uBuildDataset;

{$M+}

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, DB, DBClient, xmldom, XMLIntf, msxmldom, XMLDoc, DateUtils;

type
  TDataSetBuilder = class
  private
    { Private declarations }
    fXML: string;
    fSourceClientDataSet: TClientDataSet;
    fClientDataSet: TClientDataSet;
    fDataSetName: string;
    SourceFieldList: array of string;
    XFieldList: array of string;
    YFieldList: array of string;
    YFormulaeFieldList: array of string;
    function GetXMLNodeValue(var NodeValue: string; ParentNode: IXMLNode;
      NodeName: string): boolean;
    function BuildXIntegerValue(Formulae: string): integer;
    function BuildYIntegerValue(Formulae: string): integer;
    function BuildDateTimeValue: TDateTime;
  public
    { Public declarations }
    function Build: boolean;
    function GetDataSetData: Variant;
    function GetDataSetName: string;
    function XValueInList(Value: string): boolean;
  published
    constructor Create(SourceDataSet: TClientDataSet; XML: String);
    Destructor Destroy; override;
  end;

implementation

{ TDataSetBuilder }

// -----------------------------------------------------------------------------

constructor TDataSetBuilder.Create(SourceDataSet: TClientDataSet; XML: String);
begin
  fXML := XML;
  fSourceClientDataSet := TClientDataSet.Create(nil);
  fSourceClientDataSet.Data := SourceDataSet.Data;

  fClientDataSet := TClientDataSet.Create(nil);
end;

// -----------------------------------------------------------------------------

destructor TDataSetBuilder.Destroy;
begin
  if fClientDataSet <> nil then
  begin
    fSourceClientDataSet.Close;
    fSourceClientDataSet.Free;
  end;

  if fClientDataSet <> nil then
  begin
    fClientDataSet.Close;
    fClientDataSet.Free;
  end;

  inherited;
end;

// -----------------------------------------------------------------------------

function TDataSetBuilder.Build: boolean;
var
  XMLDoc: IXMLDocument;
  baseNode, subNode: IXMLNode;
  FieldName, FieldType: string;
  XFieldName, XFieldType, XFormulae: string;
  YValueColumnSource, YFormulae: string;
  Index, TableIndex, ColumnIndex: integer;
  RecordFound, AllowInsertUpdate: boolean;
  IntVal: integer;
  DateTimeVal: TDateTime;
  StringVal: string;
  CurrentYColumnName: string;
  CurrentYValue: integer;
begin
  Result := True;

  fClientDataSet.Close;
  fClientDataSet.FieldDefs.Clear;

  XMLDoc := TXMLDocument.Create(nil);

  XMLDoc.Active := False;
  XMLDoc.XML.Text := Trim(fXML);

  try
    try
      XMLDoc.Active := True;

      // Add all source columns
      SetLength(SourceFieldList, 0);

      for Index := 0 to fSourceClientDataSet.FieldCount - 1 do
      begin
        FieldName := fSourceClientDataSet.Fields[Index].FieldName;
        SetLength(SourceFieldList, Length(SourceFieldList) + 1);
        SourceFieldList[Length(SourceFieldList) - 1] := FieldName;
      end;

      baseNode := XMLDoc.DocumentElement.ChildNodes.FindNode('Columns');
      if baseNode <> nil then
      begin
        subNode := baseNode.ChildNodes.First;

        while subNode <> nil do
        begin
          GetXMLNodeValue(FieldName, subNode, 'Name');
          GetXMLNodeValue(FieldType, subNode, 'Type');

          if FieldType = 'integer' then
            fClientDataSet.FieldDefs.Add(FieldName, ftInteger, 0, False)
          else if FieldType = 'datetime' then
            fClientDataSet.FieldDefs.Add(FieldName, ftDateTime, 0, False)
          else if FieldType = 'string' then
            fClientDataSet.FieldDefs.Add(FieldName, ftString, 50, False);

          subNode := subNode.nextsibling;
        end;
      end;

      fClientDataSet.CreateDataSet;
      fClientDataSet.Open;

      // Get info about X column
      baseNode := XMLDoc.DocumentElement.ChildNodes.FindNode('XColumn');
      GetXMLNodeValue(XFieldName, baseNode, 'Name');
      GetXMLNodeValue(XFieldType, baseNode, 'Type');
      GetXMLNodeValue(XFormulae, baseNode, 'Formulae');

      // Build X fieldlist
      SetLength(XFieldList, 0);

      baseNode := XMLDoc.DocumentElement.ChildNodes.FindNode('XColumn')
        .ChildNodes.FindNode('Fields');

      if baseNode <> nil then
        for Index := 0 to baseNode.ChildNodes.Count - 1 do
        begin
          SetLength(XFieldList, Length(XFieldList) + 1);

          if not VarIsNull(baseNode.ChildNodes[Index].NodeValue) then
            XFieldList[Length(XFieldList) - 1] := baseNode.ChildNodes[Index]
              .NodeValue;
        end;

      // Get info about Y column
      baseNode := XMLDoc.DocumentElement.ChildNodes.FindNode('YColumn');
      GetXMLNodeValue(YValueColumnSource, baseNode, 'ColumnSource');
      GetXMLNodeValue(YFormulae, baseNode, 'Formulae');

      // Build Y Column fieldlist
      SetLength(YFieldList, 0);

      baseNode := XMLDoc.DocumentElement.ChildNodes.FindNode('YColumn');
      baseNode := baseNode.ChildNodes.FindNode('Fields');

      if baseNode <> nil then
        for Index := 0 to baseNode.ChildNodes.Count - 1 do
        begin
          SetLength(YFieldList, Length(YFieldList) + 1);

          if not VarIsNull(baseNode.ChildNodes[Index].NodeValue) then
            YFieldList[Length(YFieldList) - 1] := baseNode.ChildNodes[Index]
              .NodeValue;
        end;

      // Build Y formulae fieldlist
      SetLength(YFormulaeFieldList, 0);

      baseNode := XMLDoc.DocumentElement.ChildNodes.FindNode('YColumn');
      baseNode := baseNode.ChildNodes.FindNode('FormulaeFields');

      if baseNode <> nil then
        for Index := 0 to baseNode.ChildNodes.Count - 1 do
        begin
          SetLength(YFormulaeFieldList, Length(YFormulaeFieldList) + 1);

          if not VarIsNull(baseNode.ChildNodes[Index].NodeValue) then
            YFormulaeFieldList[Length(YFormulaeFieldList) - 1] :=
              baseNode.ChildNodes[Index].NodeValue;
        end;

      // Add records
      fSourceClientDataSet.First;

      for Index := 0 to fSourceClientDataSet.RecordCount - 1 do
      begin
        // Insert or update result table
        fClientDataSet.First;
        RecordFound := False;
        AllowInsertUpdate := True;

        IntVal := 0;
        DateTimeVal := 0;
        StringVal := '';

        // Create the (X) match value
        if XFieldType = 'integer' then
          IntVal := BuildXIntegerValue(XFormulae)
        else if XFieldType = 'datetime' then
        begin
          DateTimeVal := BuildDateTimeValue;

          if DateTimeVal = 0 then
            AllowInsertUpdate := False;
        end
        else
          StringVal := '';

        // If we have no valid datetime for X column then don't insert
        // or update the table
        if AllowInsertUpdate then
        begin
          // Find the match value in our current result records
          for TableIndex := 0 to fClientDataSet.RecordCount - 1 do
          begin
            if XFieldType = 'integer' then
            begin
              if fClientDataSet.FieldByName(XFieldName).AsInteger = IntVal then
              begin
                RecordFound := True;
                Break;
              end;
            end
            else if XFieldType = 'datetime' then
            begin
              if fClientDataSet.FieldByName(XFieldName)
                .AsDateTime = DateTimeVal then
              begin
                RecordFound := True;
                Break;
              end;
            end
            else
            begin
              if fClientDataSet.FieldByName(XFieldName)
                .AsString = StringVal then
              begin
                RecordFound := True;
                Break;
              end;
            end;

            fClientDataSet.Next;
          end;

          // Determine if we need to insert or update the table
          if not RecordFound then
          begin
            fClientDataSet.Insert;

            if XFieldType = 'integer' then
              fClientDataSet.FieldByName(XFieldName).AsInteger := IntVal
            else if XFieldType = 'datetime' then
              fClientDataSet.FieldByName(XFieldName).AsDateTime := DateTimeVal
            else
              fClientDataSet.FieldByName(XFieldName).AsString := StringVal;

            // Default columns to 0
            for ColumnIndex := Low(YFieldList) to High(YFieldList) do
              fClientDataSet.FieldByName(YFieldList[ColumnIndex])
                .AsInteger := 0;
          end
          else
            fClientDataSet.Edit;

          // Set Y value for column
          CurrentYColumnName := fSourceClientDataSet.FieldByName
            (YValueColumnSource).AsString;

          // Only update it if we have a value
          if CurrentYColumnName <> '' then
          begin
            CurrentYValue := fClientDataSet.FieldByName(CurrentYColumnName)
              .AsInteger;

            fClientDataSet.FieldByName(CurrentYColumnName).AsInteger :=
              CurrentYValue + BuildYIntegerValue(YFormulae);
          end;

          fClientDataSet.Post;
        end;

        fSourceClientDataSet.Next;
      end;

      XMLDoc.Active := False;
    except
      on E: exception do
      begin
        ShowMessage(E.Message);
        Result := False;
      end;
    end;
  finally
    XMLDoc := nil;
  end;
end;

// -----------------------------------------------------------------------------

function TDataSetBuilder.GetDataSetData: Variant;
begin
  Result := fClientDataSet.Data;
end;

// -----------------------------------------------------------------------------

function TDataSetBuilder.GetDataSetName: string;
begin
  Result := fDataSetName;
end;

// -----------------------------------------------------------------------------

function TDataSetBuilder.GetXMLNodeValue(var NodeValue: string;
  ParentNode: IXMLNode; NodeName: string): boolean;
var
  subNode: IXMLNode;
begin
  Result := False;

  if ParentNode <> nil then
  begin
    subNode := ParentNode.ChildNodes.FindNode(NodeName);

    if subNode <> nil then
    begin
      NodeValue := subNode.GetText;
      Result := True;
    end;
  end;
end;

// -----------------------------------------------------------------------------

function TDataSetBuilder.BuildXIntegerValue(Formulae: string): integer;
//var
//  MathParser: TbcParser;
//  Index, IndexMath: integer;
begin
  Result := 10;

{
  MathParser := TbcParser.Create(nil);

  try
    for Index := Low(XFieldList) to High(XFieldList) do
      if not MathParser.IsVariable(XFieldList[Index]) then
        MathParser.Variable[XFieldList[Index]] := 0;

    MathParser.Expression := Formulae;

    for IndexMath := 0 to MathParser.GetVariablesUsed.Count - 1 do
      for Index := 0 to Length(XFieldList) - 1 do
        if XFieldList[Index] = MathParser.GetVariablesUsed[IndexMath] then
        begin
          MathParser.AssignVar(XFieldList[Index],
            fSourceClientDataSet.FieldByName(XFieldList[Index]).AsInteger);
          Break;
        end;

    try
      Result := Round(MathParser.Value);
    except
      Result := 0;
    end;
  finally
    MathParser.Free;
  end;
  }
end;

// -----------------------------------------------------------------------------

function TDataSetBuilder.BuildYIntegerValue(Formulae: string): integer;
//var
//  MathParser: TbcParser;
//  Index, IndexMath: integer;
begin
  Result := 10;
  {
    MathParser := TbcParser.Create(nil);
    try
    for Index := Low(YFormulaeFieldList) to High(YFormulaeFieldList) do
    if not MathParser.IsVariable(YFormulaeFieldList[Index]) then
    MathParser.Variable[YFormulaeFieldList[Index]] := 0;

    MathParser.Expression := Formulae;

    for IndexMath := 0 to MathParser.GetVariablesUsed.Count - 1 do
    for Index := 0 to Length(YFormulaeFieldList) - 1 do
    if YFormulaeFieldList[Index] = MathParser.GetVariablesUsed
    [IndexMath] then
    begin
    MathParser.AssignVar(YFormulaeFieldList[Index],
    fSourceClientDataSet.FieldByName(YFormulaeFieldList[Index])
    .AsInteger);
    Break;
    end;

    try
    Result := Round(MathParser.Value);
    except
    Result := 0;
    end;
    finally
    MathParser.Free;
    end;
    }
end;

// -----------------------------------------------------------------------------

function TDataSetBuilder.BuildDateTimeValue: TDateTime;
var
  iYear, iMonth, iDay: integer;
  iHour, iMinute, iSecond: integer;
begin
  // Year
  if XValueInList(XFieldList[0]) then
    iYear := fSourceClientDataSet.FieldByName(XFieldList[0]).AsInteger
  else
  begin
    if XFieldList[0] <> '' then
      TryStrToInt(XFieldList[0], iYear)
    else
      iYear := 0;
  end;

  // if iYear = 0 then
  // iYear := 1970;

  // Month
  if XValueInList(XFieldList[1]) then
    iMonth := fSourceClientDataSet.FieldByName(XFieldList[1]).AsInteger
  else
  begin
    if XFieldList[1] <> '' then
      TryStrToInt(XFieldList[1], iMonth)
    else
      iMonth := 01;
  end;

  if (iMonth < 1) or (iMonth > 12) then
    iMonth := 1;

  // Day
  if XValueInList(XFieldList[2]) then
    iDay := fSourceClientDataSet.FieldByName(XFieldList[2]).AsInteger
  else
  begin
    if XFieldList[2] <> '' then
      TryStrToInt(XFieldList[2], iDay)
    else
      iDay := 01;
  end;

  if (iDay < 1) or (iDay > 31) then
    iDay := 1;

  // Hour
  if XValueInList(XFieldList[3]) then
    iHour := fSourceClientDataSet.FieldByName(XFieldList[3]).AsInteger
  else
  begin
    if XFieldList[3] <> '' then
      TryStrToInt(XFieldList[3], iHour)
    else
      iHour := 0;
  end;

  if (iHour < 0) or (iHour > 23) then
    iHour := 0;

  // Minute
  if XValueInList(XFieldList[4]) then
    iMinute := fSourceClientDataSet.FieldByName(XFieldList[4]).AsInteger
  else
  begin
    if XFieldList[4] <> '' then
      TryStrToInt(XFieldList[4], iMinute)
    else
      iMinute := 0;
  end;

  if (iMinute < 0) or (iMinute > 59) then
    iMinute := 0;

  // Second
  if XValueInList(XFieldList[5]) then
    iSecond := fSourceClientDataSet.FieldByName(XFieldList[5]).AsInteger
  else
  begin
    if XFieldList[5] <> '' then
      TryStrToInt(XFieldList[5], iSecond)
    else
      iSecond := 0;
  end;

  if (iSecond < 0) or (iSecond > 59) then
    iSecond := 0;

  try
    Result := EncodeDateTime(iYear, iMonth, iDay, iHour, iMinute, iSecond, 0);
  except
    Result := 0;
  end;
end;

// -----------------------------------------------------------------------------

function TDataSetBuilder.XValueInList(Value: string): boolean;
var
  Index: integer;
begin
  Result := False;

  for Index := 0 to Length(SourceFieldList) - 1 do
    if SourceFieldList[Index] = Value then
    begin
      Result := True;
      Break;
    end;
end;

// -----------------------------------------------------------------------------

end.
