unit uReportThread;

interface

uses
  Classes, SysUtils, Variants, xmldom, XMLIntf, msxmldom, XMLDoc,
  frxClass, frxExportPDF, frxDBSet, frxGradient, frxChBox, frxChart,
  frxCross, frxRich, frxOLE, frxBarcode, EncdDecd, DB, DBClient, Zlib, ActiveX;

type
  AdvancedReportThread = class(TThread)
  private
    { Private declarations }
  protected
    procedure Execute; override;
    procedure LoadDataset;
    procedure LoadClientDatasets;
    function BuildReport: string;
    function GetXMLNodeValue(var NodeValue: string; ParentNode: IXMLNode;
      NodeName: string): Boolean;
    function CreateDataSet(iNode: IXMLNode; ParentIndex: Integer): Integer;
    function ZDecompressString(aText: string): string;
    function ZCompressStreamToString(Data: TStream;
      aCompressionLevel: TZCompressionLevel): string;
  public
    FErrorOccurred: Boolean;
    FTemplateData: string;
    fClientData: string;
    FClientDataSets: string;
    FOutputFormat: string;
    FOutputFilename : string;

    FReportExport: string;

    constructor Create;
    destructor Destroy; override;
  end;

implementation

uses uBuildDataset;

var
  MyReport: TfrxReport;
  MyPDFExport: TfrxPDFExport;

  ReportDataSets: array of TfrxDBDataset;
  ClientDatasets: array of TClientDataSet;
  DataSources: array of TDataSource;

  { AdvancedReportThread }

  // -----------------------------------------------------------------------------

constructor AdvancedReportThread.Create;
begin
  inherited Create(True); // Don't start just yet, create suspended

  FErrorOccurred := false;
  FreeOnTerminate := True;
end;

// -----------------------------------------------------------------------------

destructor AdvancedReportThread.Destroy;
begin
  inherited;
end;

// -----------------------------------------------------------------------------

procedure AdvancedReportThread.Execute;
var
  Index: Integer;
begin
  // initialize COM library in current thread
  CoInitializeEx(nil, COINIT_MULTITHREADED);

  // create report
  MyReport := TfrxReport.Create(nil);
  // disable all messages
  MyReport.EngineOptions.SilentMode := True;
  // enable safe work in threads
  MyReport.EngineOptions.EnableThreadSafe := True;
  // don't destroy forms in threads
  MyReport.EngineOptions.DestroyForms := false;
  // disable cache
  MyReport.EngineOptions.UseFileCache := false;
  //
  MyReport.EngineOptions.ReportThread := Self;
  // disable progress bar
  MyReport.ShowProgress := false;

  MyPDFExport := TfrxPDFExport.Create(nil);
  MyPDFExport.PrintOptimized := True;
  MyPDFExport.ShowDialog := false;
  MyPDFExport.ShowProgress := false;
  MyPDFExport.OpenAfterExport := false;
  MyPDFExport.FileName := FOutputFilename;

  try
    SetLength(ReportDataSets, 0);
    SetLength(ClientDatasets, 0);
    SetLength(DataSources, 0);

    // Build main dataset
    LoadDataset;

    // Build custom datasets
    LoadClientDatasets;

    // Build report
    BuildReport;

  finally
    for Index := Low(ReportDataSets) to High(ReportDataSets) do
    begin
      ReportDataSets[Index].Close;
      ReportDataSets[Index].Free;
    end;

    for Index := Low(ClientDatasets) to High(ClientDatasets) do
    begin
      ClientDatasets[Index].Close;
      ClientDatasets[Index].Free;
    end;

    for Index := Low(DataSources) to High(DataSources) do
      DataSources[Index].Free;

    MyPDFExport.Free;
    MyReport.Free;

    SetLength(ReportDataSets, 0);
    SetLength(ClientDatasets, 0);
    SetLength(DataSources, 0);

    CoUninitialize;
  end;
end;

// -----------------------------------------------------------------------------

procedure AdvancedReportThread.LoadDataset;
var
  XML: string;
  XMLDoc: IXMLDocument;
  baseNode, detailNode: IXMLNode;
  DataSetIndex: Integer;
begin
  XML := fClientData;

  XMLDoc := TXMLDocument.Create(nil);

  XMLDoc.Active := false;
  XMLDoc.XML.Text := XML;

  try
    XMLDoc.Active := True;

    baseNode := XMLDoc.DocumentElement.ChildNodes.First;

    while baseNode <> nil do
    begin
      if baseNode.NodeName = 'dataset' then
      begin
        DataSetIndex := CreateDataSet(baseNode, -1);

        // Find and create detail datasets
        detailNode := baseNode.ChildNodes.First;

        while detailNode <> nil do
        begin
          if detailNode.NodeName = 'detail_dataset' then
            CreateDataSet(detailNode, DataSetIndex);

          detailNode := detailNode.NextSibling;
        end;
      end;
      baseNode := baseNode.NextSibling;
    end;

    XMLDoc.Active := false;
  finally
    XMLDoc := nil;
  end;
end;

// -----------------------------------------------------------------------------

procedure AdvancedReportThread.LoadClientDatasets;
var
  XMLDoc: IXMLDocument;
  baseNode: IXMLNode;
  CustomDataSet: TDataSetBuilder;
  Index, FoundIndex, DataSetIndex: Integer;
  NodeVal: string;
begin;
  XMLDoc := TXMLDocument.Create(nil);

  XMLDoc.Active := false;
  XMLDoc.XML.Text := FClientDataSets;

  try
    XMLDoc.Active := True;

    baseNode := XMLDoc.DocumentElement.ChildNodes.First;

    while baseNode <> nil do
    begin
      GetXMLNodeValue(NodeVal, baseNode, 'SourceName');

      if NodeVal <> '' then
      begin
        FoundIndex := -1;

        for Index := Low(ClientDatasets) to High(ClientDatasets) do
          if ClientDatasets[Index].Name = NodeVal then
          begin
            FoundIndex := Index;
            Break;
          end;

        if FoundIndex <> -1 then
        begin
          DataSetIndex := Length(ClientDatasets);

          CustomDataSet := TDataSetBuilder.Create(ClientDatasets[FoundIndex],
            baseNode.XML);

          CustomDataSet.Build;

          SetLength(ClientDatasets, DataSetIndex + 1);
          SetLength(DataSources, DataSetIndex + 1);

          ClientDatasets[DataSetIndex] := TClientDataSet.Create(nil);
          ClientDatasets[DataSetIndex].Name := 'ClientDataSet' +
            IntToStr(DataSetIndex);
          DataSources[DataSetIndex] := TDataSource.Create(nil);

          ClientDatasets[DataSetIndex].Data := CustomDataSet.GetDataSetData;
          DataSources[DataSetIndex].DataSet := ClientDatasets[DataSetIndex];

          // Dataset name
          GetXMLNodeValue(NodeVal, baseNode, 'Name');

          SetLength(ReportDataSets, DataSetIndex + 1);
          ReportDataSets[DataSetIndex] := TfrxDBDataset.Create(nil);
          ReportDataSets[DataSetIndex].Name := 'Dataset' +
            IntToStr(DataSetIndex);
          ReportDataSets[DataSetIndex].Tag := 1;
          ReportDataSets[DataSetIndex].UserName := NodeVal + ' custom';
          ReportDataSets[DataSetIndex].DataSource := DataSources[DataSetIndex];
          ReportDataSets[DataSetIndex].Clear;
          ReportDataSets[DataSetIndex].Initialize;

          ReportDataSets[DataSetIndex].Tag := ReportDataSets[DataSetIndex]
            .RecordCount;

          CustomDataSet.Free;
        end;
      end;

      baseNode := baseNode.NextSibling;
    end;

    XMLDoc.Active := false;
  finally
    XMLDoc := nil;
  end;
end;

// -----------------------------------------------------------------------------

function AdvancedReportThread.BuildReport: string;
var
  ReportResult: TMemoryStream;
  ReportTemplate: TStringStream;
  Index: Integer;
begin;
  Result := 'OK';

  ReportTemplate := TStringStream.Create;
  ReportResult := TMemoryStream.Create;

  try
    try
      ReportTemplate.WriteString(FTemplateData);
      ReportTemplate.Position := 0;

      if FOutputFormat = 'PDF' then
      begin
        MyReport.LoadFromStream(ReportTemplate);
        MyReport.DataSets.Clear;

        // Add datasets to report
        for Index := 0 to Length(ReportDataSets) - 1 do
          MyReport.DataSets.Add(ReportDataSets[Index]);

        if MyReport.PrepareReport() then
        begin
          MyReport.Export(MyPDFExport);
          MyReport.DataSets.Clear;
          MyReport.Clear;
        end;
      end;
    except
      on E: Exception do
        Result := 'Remote report error: ' + Trim(E.Message);
    end;
  finally
    FreeAndNil(ReportTemplate);
    FreeAndNil(ReportResult);
  end;
end;

// -----------------------------------------------------------------------------

function AdvancedReportThread.GetXMLNodeValue(var NodeValue: string;
  ParentNode: IXMLNode; NodeName: string): Boolean;
var
  subNode: IXMLNode;
begin
  Result := false;
  NodeValue := '';

  if ParentNode <> nil then
  begin
    subNode := ParentNode.ChildNodes.FindNode(NodeName);

    if subNode <> nil then
    begin
      NodeValue := subNode.GetText;
      Result := True;
    end;
  end;
end;

// -----------------------------------------------------------------------------

function AdvancedReportThread.CreateDataSet(iNode: IXMLNode;
  ParentIndex: Integer): Integer;
var
  DataSetIndex, FieldIndex: Integer;
  datasetNode, fieldNode, recordNode, recorddataNode: IXMLNode;
  MemFieldNames: array of string;
  DatasetName, DataSetIndexField, DataSetMasterField: string;
  FieldName, FieldType, FieldSize: string;
  iFieldSize: Integer;
begin
  DataSetIndex := 0;

  GetXMLNodeValue(DatasetName, iNode, 'dataset_name');
  GetXMLNodeValue(DataSetIndexField, iNode, 'dataset_index_field');
  GetXMLNodeValue(DataSetMasterField, iNode, 'dataset_master_field');

  datasetNode := iNode.ChildNodes.First;

  while datasetNode <> nil do
  begin
    // Field details
    if datasetNode.NodeName = 'fields' then
    begin
      DataSetIndex := Length(ClientDatasets);

      SetLength(ClientDatasets, DataSetIndex + 1);
      SetLength(DataSources, DataSetIndex + 1);

      ClientDatasets[DataSetIndex] := TClientDataSet.Create(nil);
      ClientDatasets[DataSetIndex].Name := 'ClientDataSet' +
        IntToStr(DataSetIndex);
      DataSources[DataSetIndex] := TDataSource.Create(nil);

      fieldNode := datasetNode.ChildNodes.First;
      SetLength(MemFieldNames, 0);

      while fieldNode <> nil do
      begin
        GetXMLNodeValue(FieldName, fieldNode, 'name');
        GetXMLNodeValue(FieldType, fieldNode, 'type');
        GetXMLNodeValue(FieldSize, fieldNode, 'size');

        SetLength(MemFieldNames, Length(MemFieldNames) + 1);
        MemFieldNames[Length(MemFieldNames) - 1] := FieldName;

        // TODO add other fieldtypes to
        {
          string
          integer
          float
          date
          time
          datetime
        }
        try
          if FieldType = 'integer' then
            ClientDatasets[DataSetIndex].FieldDefs.Add(FieldName,
              ftInteger, 0, false)
          else
          begin
            iFieldSize := StrToInt(FieldSize);

            // String fields cannot be bigger then 65530 characters
            if iFieldSize > 65530 then
              iFieldSize := 65530;

            ClientDatasets[DataSetIndex].FieldDefs.Add(FieldName, ftString,
              iFieldSize, false);
          end;
        except
        end;

        fieldNode := fieldNode.NextSibling;
      end;

      // Fields created, open dataset en set optional sort field
      if DataSetIndexField <> '' then
        ClientDatasets[DataSetIndex].IndexFieldNames := DataSetIndexField;

      ClientDatasets[DataSetIndex].CreateDataSet;
      ClientDatasets[DataSetIndex].Open;

      DataSources[DataSetIndex].DataSet := ClientDatasets[DataSetIndex];

      // Set master detail relation
      if (DataSetIndexField <> '') and (DataSetMasterField <> '') then
      begin
        ClientDatasets[DataSetIndex].MasterSource := DataSources[ParentIndex];
        ClientDatasets[DataSetIndex].IndexFieldNames := DataSetIndexField;
        ClientDatasets[DataSetIndex].MasterFields := DataSetMasterField;
      end;
    end;

    // Add Records
    if datasetNode.NodeName = 'records' then
    begin
      recordNode := datasetNode.ChildNodes.First;

      while recordNode <> nil do
      begin
        ClientDatasets[DataSetIndex].Insert;

        recorddataNode := recordNode.ChildNodes.First;
        FieldIndex := 0;

        while recorddataNode <> nil do
        begin
          if not VarIsNull(recorddataNode.NodeValue) then
            ClientDatasets[DataSetIndex].FieldByName(MemFieldNames[FieldIndex])
              .AsString := recorddataNode.NodeValue
          else
            ClientDatasets[DataSetIndex].FieldByName(MemFieldNames[FieldIndex])
              .AsString := '';

          Inc(FieldIndex, 1);
          recorddataNode := recorddataNode.NextSibling;
        end;

        ClientDatasets[DataSetIndex].Post;

        recordNode := recordNode.NextSibling;
      end;
    end;

    datasetNode := datasetNode.NextSibling;
  end;
  // while datasetNode <> nil do

  // Create fxrDataset
  if DatasetName <> '' then
  begin
    SetLength(ReportDataSets, DataSetIndex + 1);
    ReportDataSets[DataSetIndex] := TfrxDBDataset.Create(nil);
    ReportDataSets[DataSetIndex].Name := 'Dataset' + IntToStr(DataSetIndex);
    ReportDataSets[DataSetIndex].Tag := 0;
    ReportDataSets[DataSetIndex].UserName := DatasetName;
    ReportDataSets[DataSetIndex].DataSource := DataSources[DataSetIndex];
    ReportDataSets[DataSetIndex].Clear;
    ReportDataSets[DataSetIndex].Initialize;

    ReportDataSets[DataSetIndex].Tag := ReportDataSets[DataSetIndex]
      .RecordCount;
  end;

  Result := DataSetIndex;
end;

// -----------------------------------------------------------------------------

function AdvancedReportThread.ZDecompressString(aText: string): string;
var
  strInput, strOutput: TStringStream;
  Unzipper: TZDecompressionStream;
begin
  Result := '';

  strInput := TStringStream.Create(aText);
  strOutput := TStringStream.Create;
  try
    Unzipper := TZDecompressionStream.Create(strInput);
    try
      strOutput.CopyFrom(Unzipper, Unzipper.Size);
    finally
      Unzipper.Free;
    end;
    Result := strOutput.DataString;
  finally
    strInput.Free;
    strOutput.Free;
  end;
end;

// -----------------------------------------------------------------------------

function AdvancedReportThread.ZCompressStreamToString(Data: TStream;
  aCompressionLevel: TZCompressionLevel): string;
var
  strOutput: TStringStream;
  Zipper: TZCompressionStream;
begin
  Result := '';

  strOutput := TStringStream.Create;
  try
    Zipper := TZCompressionStream.Create(strOutput, aCompressionLevel);
    try
      Data.Position := 0;
      Zipper.CopyFrom(Data, Data.Size);
    finally
      Zipper.Free;
    end;
    Result := strOutput.DataString;
  finally
    strOutput.Free;
  end;
end;

end.
